from __future__ import annotations

import pandas as pd
import os


class BCLIF:
    def __init__(self, path: str, sep: str):
        self.path = path
        self._sep = sep
        self.lifparams = pd.read_csv(
            os.path.join(self.path, "lif/lifparams.csv"),
            sep=self._sep,
            header=0,
            keep_default_na=False,
        )
        self.proc = pd.read_csv(
            os.path.join(self.path, "lif/processing.csv"),
            sep=self._sep,
            header=0,
            index_col=0,
            keep_default_na=False,
        )



from __future__ import annotations


import os
import warnings
import numpy as np
import pandas as pd
from .bcfid import BCFid


class BCFTMW:
    """Storage object for CP-FTMW data

    The ``BCFTMW`` class loads FID and processing data for CP-FTMW experiments.
    It provides functions for accessing FIDs and for performing processing
    tasks that involve analyzing multiple FIDs in a single operation (e.g., sideband
    deconvolution in an LO Scan experiment).

    This class is not meant to be initialized directly by the user; it is created
    automatically by BCExperiment.


    Args:
        path: Path to fid directory
        sep: CSV delimiter

    Examples:

        Loading an experiment and reading a FID (0.csv)::

            from blackchirp import *

            exp = BCExperiment('path/to/experiment')

            # The BCFTMW object is exp.ftmw if the experiment contains CP-FTMW data.

            fid = exp.ftmw.get_fid()

        Performing upper sideband deconvolution, harmonic mean, FT range 0.1-1.0 GHz
        per segment::

            x, y = exp.ftmw.process_sideband(which='upper',
                                             avg='harmonic',
                                             min_ft_offset=100.0,
                                             max_ft_offset=1000.0)

    Attributes:

        fidparams (DataFrame): Contents of fidparams.csv file
        processing (DataFrame): Contents of proessing.csv file
        numfids (int): Number of FIDs specified in fidparams.csv

    """

    def __init__(self, path: str, sep: str):
        """Loads fidparams.csv and processing.csv

        Args:
            path: Path to fid directory
            sep: CSV delimiter
        """
        self.path = path
        self._sep = sep
        self.fidparams = pd.read_csv(
            os.path.join(self.path, "fid/fidparams.csv"),
            sep=self._sep,
            header=0,
            index_col=0,
            keep_default_na=False,
        )
        self.proc = {}
        try:
            pdf = pd.read_csv(
                os.path.join(self.path, "fid/processing.csv"),
                sep=self._sep,
                header=0,
                index_col=0,
                keep_default_na=False,
            )
            for r in pdf.itertuples():
                self.proc[r[0]] = r[1]
        except FileNotFoundError:
            pass

        self.numfids = len(self.fidparams)

    def get_fid(self, num: int = 0) -> BCFid:
        """Loads a (potentially multi-frame) FID from disk.

        For standard acquisition modes (Target Shots, Target Time, and Forever),
        the complete FID is stored as ``0.csv`` and would correspond to ``num=0``, the
        default. Any other CSV files are backups and are accessed by providing the
        desired backup number for ``num``.

        For other acquisition modes that consist of several segments (LO Scan, DR Scan),
        ``num`` corresponds to the desired segment.

        Args:
            num: Number of FID file to load.

        Returns:
            A BCFid object containing the requested FID

        Raises ValueError:
            If num is out of range.


        """
        if num >= self.numfids:
            raise ValueError(
                f"Invalid FID number ({num}). Must be between 0 and {self.numfids-1}"
            )
        return BCFid(num, self.path, self.fidparams, self._sep, self.proc)

    def process_sideband(
        self,
        which: str = "both",
        avg: str = "harmonic",
        min_ft_offset: float = None,
        max_ft_offset: float = None,
        frame: int = 0,
        verbose: bool = False,
        **proc_kwargs,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Performs sideband deconvolution on an LO Scan experiment

        See `Sideband Deconvolution <../user_guide/cp-ftmw.html#sideband-deconvolution>`_
        for an explanation of the purpose of the algorithm. The arguments correspond
        to the options available in the Sideband Processing and FID processing menus
        in Blackchirp.

        Args:
            which: {'both', 'upper', 'lower'}
                Sideband to use in deconvolution.
            avg: {'harmonic', 'geometric'}
                Whether to use a harmonic or geometric mean for overlapped segments.
            min_ft_offset: Minimum offset frequency for sideband, in MHz.
                Value is relative to the Downconversion LO frequency.
                Defaults to 0.0 MHz.
            max_ft_offset: Maximum offset frequency for sideband, in MHz.
                Value is relative to the Downconversion LO frequency.
                Defaults to the FT bandwidth.
            frame: Frame number (indexed from 0).
                If negative, all frames will be averaged
            verbose: If true, print a progress message during the deconvolution.
            \*\*proc_kwargs: FID processing settings passed as \*\*kwargs to BCFid.ft.

        Returns:
            Frequency and intensity arrays for processed spectrum.

        Raises:
            ValueError: If a provided argument is invalid.
        """

        # figure out range of data and create the global x array.
        min_probe = np.min(self.fidparams.probefreq.to_numpy())
        max_probe = np.max(self.fidparams.probefreq.to_numpy())
        fid = self.get_fid(0)
        x, ft = fid.ft(frame=frame, **proc_kwargs)
        bw = np.abs(np.max(x) - np.min(x))
        ftsp = bw / (len(x) - 1)

        si = 0
        ei = 0
        if min_ft_offset is not None:
            if min_ft_offset <= 0.0:
                raise ValueError("min_ft_offset must be positive")

            si = int(np.ceil(min_ft_offset / ftsp))
        else:
            min_ft_offset = 0.0

        if max_ft_offset is not None:
            if max_ft_offset <= 0.0:
                raise ValueError("max_ft_offset must be positive")

            ei = int(np.floor(max_ft_offset / ftsp))
        else:
            max_ft_offset = bw

        if which == "lower":
            xx = np.arange(min_probe - max_ft_offset, max_probe - min_ft_offset, ftsp)
        elif which == "upper":
            xx = np.arange(min_probe + min_ft_offset, max_probe + max_ft_offset, ftsp)
        elif which == "both":
            xx = np.arange(min_probe - max_ft_offset, max_probe + max_ft_offset, ftsp)
        else:
            raise ValueError("Which must be 'lower', 'upper', or 'both'")

        shots_array = np.zeros_like(xx)
        y_out = np.zeros_like(xx)

        if avg == "harmonic":
            avg_f = BC_harm_mean
        elif avg == "geometric":
            avg_f = BC_geo_mean
        else:
            raise ValueError("Avg must be 'geometric' or 'harmonic'")

        for i in range(self.numfids):
            if verbose:
                print(f"Processing {i+1}/{self.numfids}")
            fid = self.get_fid(i)
            if (len(fid) == 0) or (fid.shots == 0):
                continue

            if frame < 0:
                fid.average_frames()
            x, ft = fid.ft(frame=frame, **proc_kwargs)
            x = x[si:ei]
            ftu = ft.flatten()[si:ei]
            ftl = ftu[::-1]

            if fid.fidparams["sideband"] == 1:
                xl = x[::-1]
                xu = (fid.fidparams.probefreq - x) + fid.fidparams.probefreq
            else:
                t = (fid.fidparams.probefreq - x) + fid.fidparams.probefreq
                xl = t[::-1]
                xu = x

            if (which == "lower") or (which == "both"):
                yint = np.interp(xx, xl, ftl, left=0.0, right=0.0)
                yshots = np.where(yint > 0.0, fid.shots, 0.0)

                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    y_out = avg_f(y_out, yint, shots_array, yshots)
                shots_array += yshots

            if (which == "upper") or (which == "both"):
                yint = np.interp(xx, xu, ftu, left=0.0, right=0.0)
                yshots = np.where(yint > 0.0, fid.shots, 0.0)

                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    y_out = avg_f(y_out, yint, shots_array, yshots)
                shots_array += yshots

        return xx, y_out


def BC_harm_mean(
    y1: np.ndarray, y2: np.ndarray, s1: np.ndarray, s2: np.ndarray
) -> np.ndarray:
    return np.where(s1 == 0, y2, np.where(s2 == 0, y1, (s1 + s2) / (s1 / y1 + s2 / y2)))


def BC_geo_mean(
    y1: np.ndarray, y2: np.ndarray, s1: np.ndarray, s2: np.ndarray
) -> np.ndarray:
    return np.where(
        s1 == 0,
        y2,
        np.where(s2 == 0, y1, np.exp((s1 * np.log(y1) + s2 * np.log(y2)) / (s1 + s2))),
    )



# -*- coding: utf-8 -*-

from __future__ import annotations


import os
import pandas as pd
from .bcftmw import BCFTMW
from .bclif import BCLIF


class BCExperiment:
    """Container for Blackchirp experimental data

    The ``BCExperiment`` class reads in experimental data files from a Blackchirp
    Experiment, providing access to all of the data and settings associated with
    the experiment. The CP-FTMW and/or LIF data associated with the experiment are
    loaded into ``BCFTMW`` and ``BCLIF`` objects which are used to access and process
    the experimental data.

    Essentially, the ``BCExperiment`` class directly stores the data shown on the
    Header, Aux Data, and Log tabs, and its ``ftmw`` and ``lif`` properties contain
    the data shown on the CP-FTMW and LIF tabs in the Blackchirp program.

    Args:
        path: Path to data storage folder or experiment folder
        num: Experiment number (only required if path is a data storage folder)


    Raises:
        FileNotFoundError: If no ``version.csv`` file is found

    Example:
        To load data, pass the experiment number and/or the path to the
        folder containing the data. The path should point either to the base
        Blackchirp data storage folder or to the specific folder of the desired
        experiment.

        For example, if the data storage location is ``/home/user/blackchirp``, then
        to load experiment 347 (``/home/user/blackchirp/experiments/0/0/347``)::

            $ e347 = BCExperiment('/home/user/blackchirp',347)

        The experiment number is only required if pointing to a data storage folder.
        The same data could be accessed with::

            $ e347 = BCExperiment('/home/user/blackchirp/experiments/0/0/347')

        If, instead, folder 347 has been copied into the same directory as your python
        script, the experiment can be accessed with::

            $ e347 = BCExperiment('./347')

        Each csv file in the Experiment folder is loaded as a
        `pandas DataFrame <https://pandas.pydata.org/pandas-docs/stable/reference/frame.html>`_.
        These DataFrames are each stored as a variable with the same name as the csv
        file. For example, to see the contents of hardware.csv::

            $ e437.hardware

                            key   subKey
            0             AWG.0  virtual
            1           Clock.0    fixed
            2           Clock.1    fixed
            3   FtmwDigitizer.0  virtual
            4  PulseGenerator.0  virtual


    Attributes:
        num (int): Experiment number
        path (str): Experiment path
        version (pd.DataFrame): Contents of version.csv.
            Information about the Blackchirp version used to acquire the data
        header (pd.DataFrame): Contents of header.csv.
            General experimental parameters.
        objectives (pd.DataFrame): Contents of objectives.csv.
            Information about the goals of the experiment (FTMW type, LIF, etc)
        log (pd.DataFrame): Contents of log.csv.
            All messages printed to Blackchirp's log during the experiment.
        hardware (pd.DataFrame): Contents of hardware.csv.
            All hardware items and their implementation keys.
        clocks (pd.DataFrame): Contents of clocks.csv
            Configurations of all clocks at each experimental step.
        auxdata (pd.DataFrame, optional): Contents of auxdata.csv (if present).
            Data shown on `Aux Data <user_guide/rolling-aux-data.html>`_ plots
            during the experiment.
        chirps (pd.DataFrame, optional): Contents of chirps.csv (if present).
            Details of all chirps associated with a CP-FTMW acquisition.
        ftmw (BCFTMW, optional): Contents of fid directory.
            This object provides an interface for accessing CP-FTMW data.
        lif (BCLIF, optional): Contents of lif directory.
            This object provides an interface for accessing LIF data.


    """

    def __init__(self, path: str = ".", num: int = None):
        if num is not None:
            self._mil = num // 1000000
            self._th = num // 1000
            self.num = num
        self.path = path
        self._sep = ";"

        if os.path.exists(os.path.join(path, "version.csv")):
            self.path = path
        elif num is not None:  # maybe this is a savePath
            testpath = os.path.join(
                path, f"experiments/{self._mil}/{self._th}/{self.num}"
            )
            if os.path.exists(os.path.join(testpath, "version.csv")):
                self.path = testpath
            else:
                raise FileNotFoundError(f"Could not find Blackchirp data at {testpath}")
        else:
            raise FileNotFoundError(
                f"Could not find Blackchirp data at {os.path.abspath(path)}"
            )

        # read separator character from version file
        with open(os.path.join(self.path, "version.csv"), "r") as v:
            l = v.readline()
            self._sep = l.strip()
        self.version = pd.read_csv(
            os.path.join(self.path, "version.csv"),
            sep=self._sep,
            header=1,
            keep_default_na=False,
        )

        self.header = pd.read_csv(
            os.path.join(self.path, "header.csv"),
            sep=self._sep,
            header=0,
            dtype={
                "ObjKey": str,
                "ArrayKey": str,
                "ArrayIndex": "Int64",
                "ValueKey": str,
                "Value": str,
                "Units": str,
            },
            keep_default_na=False,
        )
        self.num = int(
            self.header.query(
                "ObjKey == 'Experiment' and ValueKey == 'Number'"
            ).Value.iloc[0]
        )

        self.objectives = pd.read_csv(
            os.path.join(self.path, "objectives.csv"),
            sep=self._sep,
            header=0,
            keep_default_na=False,
        )
        self.log = pd.read_csv(
            os.path.join(self.path, "log.csv"),
            sep=self._sep,
            header=0,
            keep_default_na=False,
        )
        self.hardware = pd.read_csv(
            os.path.join(self.path, "hardware.csv"),
            sep=self._sep,
            header=0,
            keep_default_na=False,
        )
        self.clocks = pd.read_csv(
            os.path.join(self.path, "clocks.csv"),
            sep=self._sep,
            header=0,
            keep_default_na=False,
        )

        try:
            self.auxdata = pd.read_csv(
                os.path.join(self.path, "auxdata.csv"),
                sep=self._sep,
                header=0,
                keep_default_na=False,
            )
        except FileNotFoundError:
            pass

        try:
            self.chirps = pd.read_csv(
                os.path.join(self.path, "chirps.csv"),
                sep=self._sep,
                header=0,
                keep_default_na=False,
            )
        except FileNotFoundError:
            pass

        if os.path.exists(os.path.join(self.path, "fid")):
            self.ftmw = BCFTMW(self.path, self._sep)

        if os.path.exists(os.path.join(self.path, "lif")):
            self.lif = BCLIF(self.path, self._sep)

    def header_unique_keys(self) -> set[str]:
        """Fetch all unique ObjKeys in experiment header

        Returns:
            List of unique header keys

        """

        return set(self.header.ObjKey.tolist())

    def header_rows(
        self, objKey: str = None, valKey: str = None, arrKey: str = None
    ) -> pd.DataFrame:
        """Fetch rows from the header file matching conditions

        Filters rows in the header according to ObjKey, ValueKey, and ArrayKey.
        Any combination of these (or none) may be specified to filter.

        Args:
            objKey: Object key in header
            valKey: Value key in header
            arrKey: Array key in header

        Returns:
            DataFrame with matching wors

        """

        df = self.header
        if objKey is not None:
            df = df.query(f"ObjKey == '{objKey}'")
        if valKey is not None:
            df = df.query(f"ValueKey == '{valKey}'")
        if arrKey is not None:
            df = df.query(f"ArrayKey == '{arrKey}'")

        return df

    def header_value(
        self, objKey: str, valKey: str, idx: int = 0, arrKey: str = None
    ) -> str:
        """Fetch one value from header

        The objKey and valKey (and arrKey, if specified) are used to filter the header.
        Then the idx value is used to determine which row to return. If a match is not
        found, an empty string is returned

        Args:
            objKey: Object key in header
            valKey: Value key in header
            idx: Row number to return (optional)
            arrKey: Array key in header (optional)

        Returns:
            Matching value or empty string
        """

        df = self.header_rows(objKey, valKey, arrKey)
        if len(df) <= idx:
            return ""

        return df.Value.iloc[idx]

    def header_unit(
        self, objKey: str, valKey: str, idx: int = 0, arrKey: str = None
    ) -> str:
        """Fetch one unit value from header

        The objKey and valKey (and arrKey, if specified) are used to filter the header.
        Then the idx value is used to determine which row to return. If a match is not
        found, an empty string is returned

        Args:
            objKey: Object key in header
            valKey: Value key in header
            idx: Row number to return (optional)
            arrKey: Array key in header (optional)

        Returns:
            Matching value or empty string
        """

        df = self.header_rows(objKey, valKey, arrKey)
        if len(df) <= idx:
            return ""

        out = df.Units.iloc[idx]
        if out != out:
            return ""

        return out
